/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.yuandian.dataflow.statemachine;

import org.apache.ratis.conf.RaftProperties;
// import org.apache.ratis.examples.common.Constants; 
import org.apache.ratis.grpc.GrpcConfigKeys;
import org.apache.ratis.protocol.RaftGroup;
import org.apache.ratis.protocol.RaftGroupId;
import org.apache.ratis.protocol.RaftPeer;
import org.apache.ratis.server.RaftServer;
import org.apache.ratis.server.RaftServerConfigKeys;
import org.apache.ratis.util.NetUtils;

import io.netty.util.internal.StringUtil;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Scanner;
import java.util.UUID;

import static java.nio.charset.StandardCharsets.UTF_8;

/**
 * Simplest Ratis server, use a simple state machine {@link StateMachine}
 * which maintain a counter across multi server.
 * This server application designed to run several times with different
 * parameters (1,2 or 3). server addresses hard coded in {@link Constants}
 * <p>
 * Run this application three times with three different parameter set-up a
 * ratis cluster which maintain a counter value replicated in each server memory
 */
public final class StateServer implements Closeable {
  private final RaftServer server;

  public static final UUID CLUSTER_GROUP_ID = UUID.fromString("79642d72-6166-742d-6461-7461666c6f77");

  public StateServer(RaftPeer peer,  ArrayList<RaftPeer> peers, File storageDir) throws IOException {
    //create a property object
    RaftProperties properties = new RaftProperties();

    //set the storage directory (different for each peer) in RaftProperty object
    RaftServerConfigKeys.setStorageDir(properties, Collections.singletonList(storageDir));

    //set the port which server listen to in RaftProperty object
    final int port = NetUtils.createSocketAddr(peer.getAddress()).getPort();
    GrpcConfigKeys.Server.setPort(properties, port);

    //create the counter state machine which hold the counter value
    StateMachine counterStateMachine = new StateMachine();

    RaftGroup raftGroup = RaftGroup.valueOf(
      RaftGroupId.valueOf(CLUSTER_GROUP_ID), peers);
      
    //create and start the Raft server
    this.server = RaftServer.newBuilder()
        .setGroup(raftGroup)
        .setProperties(properties)
        .setServerId(peer.getId())
        .setStateMachine(counterStateMachine)
        .build();
  }

  public void start() throws IOException {
    server.start();
  }

  @Override
  public void close() throws IOException {
    server.close();
  }

  public static void main(String[] args) throws IOException {
    if (args.length < 1) {
      System.err.println("Usage: java -cp *.jar org.apache.ratis.examples.counter.server.StateServer {serverIndex}");
      System.err.println("{serverIndex} could be 1, 2 or 3");
      System.exit(1);
    }

    var peers = new ArrayList<RaftPeer>();
    String[] addresses = new String[]{"localhost:4440","localhost:4441","localhost:4442"};

    for (int i = 0; i < addresses.length; i++) {
      var port = addresses[i].split(":")[1];
      peers.add(RaftPeer.newBuilder().setId("" + port).setAddress(addresses[i]).build());
    }

    //find current peer object based on application parameter
    final RaftPeer currentPeer = peers.get(Integer.parseInt(args[0]));

    //start a counter server
    final File storageDir = new File("./raftdata/" + currentPeer.getId());
    final StateServer stateServer = new StateServer(currentPeer, peers, storageDir);
    
    stateServer.start();
 
    


    //exit when any input entered
    Scanner scanner = new Scanner(System.in, UTF_8.name());
    scanner.nextLine();
    stateServer.close();
  }
}
